#include <cmath>
#include <iostream>
#include <vector>

#include <GL/glew.h>
#include <GL/freeglut.h>
#include <GL/freeglut_ext.h>

#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>

#include "moglclass.h"
#include "shaders.h"
#include "vertexBufferObject.h"

using namespace moglclass;

#define NUMSHADERS 2
Shader shShaders[NUMSHADERS];
Program spMain;

// Cube vertices
std::vector<glm::vec4> cubeVertices;
// Cube indices
std::vector<int> cubeIndices;
// Cube colors
std::vector<glm::vec4> cubeColors;
// Cube VAO
GLuint cubeVAO;


// wireframe/filled
bool g_wiremode = true;

//int uniformMVP;
glm::mat4 anim(1.0f);

////////////////////////////////////////////////////////////////////////////////////////////////////

GLuint SetData(float size = 0.7f)
{
	// Vertice Positions
	cubeVertices.push_back(glm::vec4(-size, -size, -size, 1.0f));	// 0
	cubeVertices.push_back(glm::vec4( size, -size, -size, 1.0f));	// 1
	cubeVertices.push_back(glm::vec4( size,  size, -size, 1.0f));	// 2
	cubeVertices.push_back(glm::vec4(-size,  size, -size, 1.0f));	// 3
	cubeVertices.push_back(glm::vec4(-size, -size,  size, 1.0f));	// 4
	cubeVertices.push_back(glm::vec4( size, -size,  size, 1.0f));	// 5
	cubeVertices.push_back(glm::vec4( size,  size,  size, 1.0f));	// 6
	cubeVertices.push_back(glm::vec4(-size,  size,  size, 1.0f));	// 7

	// Vertices Colors
	cubeColors.push_back(glm::vec4(1.0f, 0.0f, 0.0f, 1.0f));
	cubeColors.push_back(glm::vec4(0.0f, 1.0f, 0.0f, 1.0f));
	cubeColors.push_back(glm::vec4(0.0f, 0.0f, 1.0f, 1.0f));
	cubeColors.push_back(glm::vec4(0.0f, 0.0f, 0.0f, 1.0f));
	cubeColors.push_back(glm::vec4(1.0f, 0.0f, 0.0f, 1.0f));
	cubeColors.push_back(glm::vec4(0.0f, 1.0f, 0.0f, 1.0f));
	cubeColors.push_back(glm::vec4(0.0f, 0.0f, 1.0f, 1.0f));
	cubeColors.push_back(glm::vec4(0.0f, 0.0f, 0.0f, 1.0f));

	// Vertice Indices
	cubeIndices.push_back(4);	// front
	cubeIndices.push_back(5);
	cubeIndices.push_back(6);
	cubeIndices.push_back(4);
	cubeIndices.push_back(6);
	cubeIndices.push_back(7);

	cubeIndices.push_back(1);	// back
	cubeIndices.push_back(0);
	cubeIndices.push_back(3);
	cubeIndices.push_back(1);
	cubeIndices.push_back(3);
	cubeIndices.push_back(2);

	cubeIndices.push_back(0);	// left
	cubeIndices.push_back(4);
	cubeIndices.push_back(7);
	cubeIndices.push_back(0);
	cubeIndices.push_back(7);
	cubeIndices.push_back(3);

	cubeIndices.push_back(5);	// right
	cubeIndices.push_back(1);
	cubeIndices.push_back(2);
	cubeIndices.push_back(5);
	cubeIndices.push_back(2);
	cubeIndices.push_back(6);

	cubeIndices.push_back(7);	// top
	cubeIndices.push_back(6);
	cubeIndices.push_back(2);
	cubeIndices.push_back(7);
	cubeIndices.push_back(2);
	cubeIndices.push_back(3);

	cubeIndices.push_back(0);	// bottom
	cubeIndices.push_back(1);
	cubeIndices.push_back(5);
	cubeIndices.push_back(0);
	cubeIndices.push_back(5);
	cubeIndices.push_back(4);

	// VAO
	GLuint vao;
	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);

	// VBO
	GLuint vbo[2], ibo;
	glGenBuffers(2, &vbo[0]);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
	glBufferData(GL_ARRAY_BUFFER, 8*sizeof(glm::vec4), &cubeVertices[0], GL_STATIC_DRAW);
	glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
	glBufferData(GL_ARRAY_BUFFER, 8*sizeof(glm::vec4), &cubeColors[0], GL_STATIC_DRAW);
	glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(1);

	// IBO
	glGenBuffers(1, &ibo);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 36*sizeof(int), &cubeIndices[0], GL_STATIC_DRAW);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo);

	glBindVertexArray(0);
	cubeVertices.clear();
	cubeColors.clear();
	cubeIndices.clear();

	return vao;
}

void init( void )
{
    // Load shaders and use the resulting shader program
	spMain.createProgram();
	shShaders[0].loadShader("simple3.vs", GL_VERTEX_SHADER);
	shShaders[1].loadShader("simple2.fs", GL_FRAGMENT_SHADER);
	spMain.attachShaderToProgram(&shShaders[0]);
	spMain.attachShaderToProgram(&shShaders[1]);
	spMain.linkProgram();
	spMain.useProgram();

	// Set data
	cubeVAO = SetData();

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	glClearColor( 1.0, 1.0, 1.0, 1.0 ); // white background
}

void display( void )
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );     // clear the window

	glm::mat4 model = glm::translate(glm::mat4(1.0f), glm::vec3(0.0, 0.0, 0.0)); 
	glm::mat4 view = glm::lookAt(glm::vec3(0.0, 2.0, 5.0), glm::vec3(0.0, 0.0, 0.0), glm::vec3(0.0, 1.0, 0.0)); 
	glm::mat4 projection = glm::perspective(45.0f, 1.0f, 0.1f, 100.0f); 
	glm::mat4 mvp = projection * view * model * anim;  
	spMain.useProgram();
	spMain.setUniform("gMVP", mvp); 

		// set drawmode
	if (g_wiremode) 
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);  // draw wireframe 
	else
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);	// draw filled 

	// cube
	glBindVertexArray(cubeVAO);
	glDrawElements(GL_TRIANGLES, 36, GL_UNSIGNED_INT, 0);       // draw filled
	
	glBindVertexArray(0);
	glDisableVertexAttribArray(0);

    glutSwapBuffers();
}

void update() 
{ 
	float angle = glutGet(GLUT_ELAPSED_TIME) / 100000.0 * 45;  // 45 per second 
	glm::vec3 axis_y(0, 1, 0); 
	anim = glm::rotate(glm::mat4(1.0f), angle, axis_y);  

	glutPostRedisplay(); 
} 

void keyboard( unsigned char key, int x, int y )
{
	switch (key)  
	{    
	case 0x1B:    
	case 'q':    
	case 'Q':      
		exit(0);     
	case ' ':
		g_wiremode = !g_wiremode;
		break;  
	}
	glutPostRedisplay();
}

void cleanup()
{
	spMain.deleteProgram();
}

int main( int argc, char **argv )
{
	glutInit(&argc, argv);
    glutInitDisplayMode( GLUT_RGBA|GLUT_DOUBLE|GLUT_DEPTH );
    glutInitWindowSize( 512, 512 );
    glutCreateWindow(argv[0]);
    
	glewInit();
    init();

    glutDisplayFunc( display );
    glutKeyboardFunc( keyboard );
	glutIdleFunc( update );

    glutMainLoop();

	cleanup();

    return 0;
}
